/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ws.ladder.util.ObjectCreator;

public class ChangeSummaryImpl implements ChangeSummary {

	private DataObject root;
	private boolean logging;
	private Set<DataObject> created = new HashSet<DataObject>();
	private Set<DataObject> deleted = new HashSet<DataObject>();
	private Set<DataObject> modified = new HashSet<DataObject>();
	private Map<DataObject, DataObjectOwner> oldOwner = new HashMap<DataObject, DataObjectOwner>();
	private Map<DataObject, Map<String, Setting>> oldValues = new HashMap<DataObject, Map<String, Setting>>();
	
	public class DataObjectOwner {
		private DataObject owner;
		private Property property;
		
		public DataObjectOwner(DataObject owner, Property property) {
			this.owner = owner;
			this.property = property;
		}
		
		public DataObject getOwner() {
			return owner;
		}
		
		public Property getProperty() {
			return property;
		}
	}
	
	public class SettingImpl implements Setting {
		
		Property property;
		Object   value;
		boolean  isSet;

		public SettingImpl(Property property, Object value) {
			this.property = property;
			this.value = value;
			isSet = true;
		}
		
		public void unset() {
			isSet = false;
		}
		
		@Override
		public Property getProperty() {
			return property;
		}

		@Override
		public Object getValue() {
			return value;
		}

		@Override
		public boolean isSet() {
			return isSet;
		}
		
	}

	public ChangeSummaryImpl(DataObject root) {
		this.root = root;
	}
	
	@Override
	public boolean isLogging() {
		return logging;
	}

	@Override
	public DataGraph getDataGraph() {
		return null;
	}

	/**
	 * Union of objects returned by the following methods:
	 *  isCreated(DataObject)
	 *  isDeleted(DataObject)
	 *  isModified(DataObject)
	 */
	@Override
	public List<?> getChangedDataObjects() {
		List<DataObject> result = new ArrayList<DataObject>();

		result.addAll(created);
		result.addAll(deleted);
		result.addAll(modified);

		return result;
	}

	@Override
	public boolean isCreated(DataObject dataObject) {
		return created.contains(dataObject);
	}

	@Override
	public boolean isDeleted(DataObject dataObject) {
		return deleted.contains(dataObject);
	}

	@Override
	public List<?> getOldValues(DataObject dataObject) {
		return new ArrayList<Setting>(oldValues.get(dataObject).values());
	}

	@Override
	public void beginLogging() {
		logging = true;
	}

	@Override
	public void endLogging() {
		logging = false;
	}

	@Override
	public boolean isModified(DataObject dataObject) {
		return modified.contains(dataObject);
	}

	@Override
	public DataObject getRootObject() {
		return root;
	}

	@Override
	public Setting getOldValue(DataObject dataObject, Property property) {
		return oldValues.get(dataObject).get(property.getName());
	}

	@Override
	public DataObject getOldContainer(DataObject dataObject) {
		return oldOwner.get(dataObject).getOwner();
	}

	@Override
	public Property getOldContainmentProperty(DataObject dataObject) {
		return oldOwner.get(dataObject).getProperty();
	}

	@Override
	public Sequence getOldSequence(DataObject dataObject) {
		return null;
	}

	@Override
	public void undoChanges() {
		// Go through the list of modified properties and set them back to their original values
		for(DataObject modified : oldValues.keySet()) {
			Map<String, Setting> originalValues = oldValues.get(modified);
			for(String property: originalValues.keySet()) {
				Setting original = originalValues.get(property);
				((ExtendedProperty)original.getProperty()).setValue(modified, original.getValue());
			}
		}
	}

	public void markDeleted(DataObject dataObject, DataObject oldContainer, Property oldContainmentProperty) {
		// The deleted object should be removed from its container after the container's value has been snapshot
		
		deleted.add(dataObject);
		markModified(oldContainer, oldContainmentProperty, ((ExtendedProperty)oldContainmentProperty).getValue(dataObject));
	}
	
	public void markCreated(DataObject dataObject, DataObject container, Property containmentProperty) {
		// The new object should be added to its container after the container's value has been snapshot
		
		created.add(dataObject);
		markModified(container, containmentProperty, ((ExtendedProperty)containmentProperty).getValue(dataObject));
	}
	
	public void markModified(DataObject dataObject, Property property, Object oldValue) {
		if(!logging)
			return;
		
		Map<String, Setting> oldValueMap = oldValues.get(dataObject);
		
		if(oldValueMap != null && oldValueMap.get(property.getName()) != null)
			return; // We already have a copy of the original value
		
		ObjectCreator objectCreator = ((ExtendedDataObject)getRootObject()).getObjectCreator();
		Setting setting = new SettingImpl(property, ((ExtendedProperty)property).shallowCopy(dataObject, objectCreator));
		if(oldValueMap == null)
			oldValueMap = new HashMap<String, Setting>();
		
		oldValueMap.put(property.getName(), setting);
		modified.add(dataObject);
	}
}
