/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.IdentityHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import ws.ladder.model.action.CollectionUpdateAction;
import ws.ladder.model.action.ElementAction;
import ws.ladder.model.action.Executable;
import ws.ladder.model.action.MigratorActionFactory;
import ws.ladder.model.action.PropertyKey;
import ws.ladder.model.action.SetterAction;
import ws.ladder.service.PersistenceOrchestrator;
import ws.ladder.util.ClassUtil;
import ws.ladder.util.IdentityArrayList;
import ws.ladder.util.ObjectCreator;

@Component
@Scope(value = "prototype")
public class DependencyGraph {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());
	public  static final String AGGREGATE_NAME_PREFIX = "AGGREGATE";

	private PersistenceOrchestrator persistenceOrchestrator;
	private ObjectCreator           objectCreator;
	private Map<String, Node>       aggregateRoots = new HashMap<String, Node>();  // transient or modified roots	
	private IdentityArrayList<Node> transientEntities = new IdentityArrayList<Node>();
	private Set<Edge>               persistentLinks = new HashSet<Edge>(); 
	private Set<Edge>               cascadedLinks = new HashSet<Edge>();	
	private Map<PropertyKey, List<Executable>> propertyActions = new HashMap<PropertyKey, List<Executable>>(); // Includes add, remove and reposition collection elements

	public DependencyGraph(ObjectCreator objectCreator) {
		this.objectCreator = objectCreator;
		this.persistenceOrchestrator = objectCreator.getPersistenceOrchestrator();
	}

	private void getPotentialRoots(CallInfo ci) {
		int i = 1;
		for(ExtendedDataObject obj: objectCreator.getDataObjects()) {
			if(!obj.isReferenceType()) // Only reference types represent persistent entities
				continue;

			Node node = (Node) obj;
			// The aggregate root should either be the target root of this operation or be reachable from the target root 
			// Also the aggregate should either be transient or have been modified 
			if( (!node.isPersistent() || node.isAggregateModified()) && (!node.isDependent() && (node.getIncoming().size() > 0 || node == ci.getOutput())) )
				aggregateRoots.put(AGGREGATE_NAME_PREFIX+i++, node);

			if(!node.isPersistent() && obj.isEntity())
				transientEntities.add(node);
		}
	}

	private void findRealRoots() {
		Set<String> aggregateNames = new HashSet<String>(aggregateRoots.keySet());

		for(Map.Entry<String, Node> entry: aggregateRoots.entrySet()) {
			if(entry.getValue().getAggregateName() == null)
				entry.getValue().setAggregateSeal(entry.getKey());
		}

		// check to see which are not real aggregate roots
		for(String aggregateName: aggregateNames) {
			Node node = aggregateRoots.get(aggregateName);
			if(!node.getAggregateName().equals(aggregateName))
				aggregateRoots.remove(aggregateName);
		}
	}

	private void breakCascadedLinks(List<Node> transientNodes) {

		for(Node node: transientNodes) {
			for(Edge edge: node.getOutgoing()) {			
				if(edge.isCascaded())
					cascadedLinks.add(edge);
			}
		}

		for(Edge edge: cascadedLinks)
			edge.disconnect();		

		logger.debug("Number of broken cascaded links: " + cascadedLinks.size());
	}		

	private void breakPersistentLinks(List<Node> transientNodes) {

		for(Node node: transientNodes) {
			for(Edge edge: node.getOutgoing()) {			
				if(edge.getTarget().isPersistent())
					persistentLinks.add(edge);
			}
		}

		for(Edge edge: persistentLinks)
			edge.disconnect();

		logger.debug("Number of broken persistent links: " + persistentLinks.size());
	}		

	private void fixCascadedLinks() {
		// Once the tree is processed, re-link the broken links
		for(Edge edge: cascadedLinks)
			edge.connect();
	}	

	private void fixPersistentLinks() {
		// Once the tree is processed, re-link the broken links
		for(Edge edge: persistentLinks)
			edge.connect();
	}		

	/**
	 * In preSort we break the following types of links
	 * 1. cascade links from the transient instances
	 * 2. links from transient to existing persistent instances
	 * optionally we can break the non-required links
	 */
	private void preSave(List<Node> transientNodes) {
		breakCascadedLinks(transientNodes);
		breakPersistentLinks(transientNodes);
	}

	private List<Node> getTransient() {
		List<Node> result;
		if(persistenceOrchestrator.canProcessAggregate())
			result = getTransient(aggregateRoots.values());
		else
			result = transientEntities;

		return result;
	}

	private void persistRoots() {
		List<Node> sortOrder;

		List<Node> transientNodes = getTransient();
		preSave(transientNodes);

		sortOrder = toposort(transientNodes);
		fixPersistentLinks(); // Persistent links needed for save

		for(Node node: sortOrder) {
			persistenceOrchestrator.saveOrUpdate(node.getInstance());
			node.setModified(false);
		}

		fixCascadedLinks(); // Now all the transient instances have been saved, so restore all the links back	

		// save the modified roots
		for(Node root: aggregateRoots.values())
			persistenceOrchestrator.saveOrUpdate(root.getInstance());
	}

	private List<Node> getTransient(Collection<Node> nodes) {
		List<Node> result = new ArrayList<Node>();
		for(Node node: nodes) 
			if(!node.isPersistent())
				result.add(node);

		return result;
	}

	private List<Node> getRoots(Collection<Node> transientObjects) {
		List<Node> result = new ArrayList<Node>();

		next: for(Node node: transientObjects) {
			if(node.getIncoming().size() == 0) {
				result.add(node);
				continue;
			}

			// If it has incoming edge check that they are only from persistent instances
			for(Edge incomingEdge: node.getIncoming()) {
				if(incomingEdge.fromTransient())
					continue next;
			}
			result.add(node);
		}

		return result;
	}
	
	private Node getUnMarked(Collection<Node> transientObjects) {
		Node result = null;
		for(Node n: transientObjects) {
			if(!n.isVisited()) {
				result = n;
				break;
			}
		}
		
		return result;
	}
	

	/**
	 * Topological sort algorithm by Tarjan
	 * @param transientObjects
	 * @return
	 */
	private List<Node> toposort(Collection<Node> transientObjects) {
		for(Node t: transientObjects)
			t.clearVisited();
		
		List<Node> sortedNodes = new ArrayList<Node>();
		Node unmarked = getUnMarked(transientObjects);
		while(unmarked != null) {
			visit(unmarked, new IdentityHashMap<Node, Object>(), sortedNodes);
			unmarked = getUnMarked(transientObjects);
		}
		
		return sortedNodes;
	}

	private void visit(Node node, Map<Node, Object> tempMarked, List<Node> sortedNodes) {
		if(tempMarked.containsKey(node))
			throw new RuntimeException("Cycle detected, the aggregate is not a DAG");
		
		if(!node.isVisited()) {
			tempMarked.put(node, null);
			
			for(Edge outgoingEdge: node.getOutgoing()) {
				if(outgoingEdge.toTransient() && outgoingEdge.isRequired())
					visit(outgoingEdge.getTarget(), tempMarked, sortedNodes);
			}
			if( !((ExtendedDataObject)node).getType().isDataType()) {
				node.setVisited(true);
				sortedNodes.add(node);
			}
		} 
	}

	public void persistGraph(CallInfo callInfo) {
		getPotentialRoots(callInfo);
		findRealRoots();
		persistRoots();
	}

	public void processActions(Settings settings) {
		// Process the uni-directional actions
		Set<PropertyKey> uniDirKeys = new HashSet<PropertyKey>();
		for(PropertyKey key: propertyActions.keySet()) {
			if( !((ExtendedProperty)key.getProperty()).isBiDirectional() )
				uniDirKeys.add(key);
		}
		for(PropertyKey uniDirKey: uniDirKeys) {
			List<Executable> actions = propertyActions.remove(uniDirKey);
			processActions(uniDirKey, actions);
		}

		settings.getInterceptor().preBiDirActionStage( Collections.unmodifiableMap(propertyActions) );

		for(Map.Entry<PropertyKey, List<Executable>> entry: propertyActions.entrySet())
			processActions(entry.getKey(), propertyActions.get(entry.getKey()));
	}

	private void processActions(PropertyKey key, List<Executable> actions) {
		if(key.getProperty().isMany()) {
			for(Executable action: actions) {
				action.execute();
			}
		} else
			processToOne(actions);
	}

	public void processToOne(List<Executable> actions) {
		// We follow the system where we execute the action that has a non-null value if one is present or execute the null setting action
		// This policy is present to automatically fix broken links
		// The other policy we can have (maybe through configuration) is to throw an exception if a conflict is detected.
		// A conflict can occur if two actions set different values. It is still desirable to throw an exception if two different non-null values are being set.

		SetterAction actionToExecute = null;
		for(Executable action: actions) {
			SetterAction setterAction = (SetterAction) action;
			if(actionToExecute == null)
				actionToExecute = setterAction;
			if( ClassUtil.getInstance(setterAction.getValue()) != null) {
				if(ClassUtil.getInstance(actionToExecute.getValue()) != null && ClassUtil.getInstance(actionToExecute.getValue()) != ClassUtil.getInstance(setterAction.getValue()) )
					throw new IllegalArgumentException("Two different objects cannot share a reference to the same object in a ToOne bi-directional relationship.");
				else
					actionToExecute = setterAction;

			}
		}
		actionToExecute.execute();		
	}

	/**
	 * If this is a element action, the migrator action is found and the element action is recorded on it
	 * @param action
	 */
	public void addAction(Executable action) {

		List<Executable> actions = getActions(action.getKey());

		if (ElementAction.class.isAssignableFrom(action.getClass())) {
			CollectionUpdateAction collectionAction = getOrCreateMigratorAction(action.getKey());
			collectionAction.addTriggeredByOppositeAction((ElementAction) action);
		} else
			actions.add(action);
	}

	private List<Executable> getActions(PropertyKey key) {
		List<Executable> actions = propertyActions.get(key);
		if(actions == null) {
			actions = new ArrayList<Executable>();
			propertyActions.put((PropertyKey) key, actions);
		}		

		return actions;
	}

	public CollectionUpdateAction getOrCreateMigratorAction(PropertyKey key) {
		// actions should not be null
		List<Executable> actions = getActions(key);

		CollectionUpdateAction migrator = null;		
		if(actions.size() == 0) {
			migrator = MigratorActionFactory.getInstance(key);
			addAction(migrator);
			return migrator;
		}

		if(actions.size() == 1) {
			migrator = (CollectionUpdateAction) actions.get(0);
		} else
			throw new IllegalStateException("Collection property should have only one migrator action");

		return migrator;
	}

}