/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.persistence.AccessType;
import javax.persistence.metamodel.Attribute.PersistentAttributeType;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.model.event.CollectionEvent;
import ws.ladder.util.ObjectCreator;


public class DynamicProperty implements ExtendedProperty {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	protected ExtendedProperty containmentProperty;
	protected String           name;
	private   DynamicType      type;

	public DynamicProperty(String name, DynamicType dynamicType, ExtendedProperty containmentProperty) {
		this.name = name;
		this.type = dynamicType;
		this.containmentProperty = containmentProperty;
	}

	@Override
	public Type getType() {
		return type;
	}

	@Override
	public List<?> getAliasNames() {
		return new ArrayList<String>();
	}	

	@Override
	public Type getKeyType() {
		return null;
	}

	@Override
	public Type getElementType() {
		return null;
	}	

	@Override
	public Property getOpposite() {
		return containmentProperty.getOpposite();
	}

	@Override
	public Property getMappedBy() {
		return containmentProperty.getMappedBy();
	}

	@Override
	public Property getMapOf() {
		return containmentProperty.getMapOf();
	}	

	@Override
	public Type getContainingType() {
		return containmentProperty.getContainingType();
	}	

	@Override
	public void setMapKeyOf(ExtendedProperty property) {
		throw new UnsupportedOperationException("Cannot set the mapKey for a dynamic property");
	}

	@Override
	public void setIndexOf(ExtendedProperty property) {
		throw new UnsupportedOperationException("Cannot set the indexOf for a dynamic property");
	}	

	@Override
	public void initPositionProperty() {
		throw new UnsupportedOperationException("Cannot set the positionProperty for a dynamic property");
	}	

	@Override
	public Property getPositionProperty() {
		return containmentProperty.getPositionProperty();
	}

	@Override
	public AccessType getAccessType() {
		return containmentProperty.getAccessType();
	}

	@Override
	public boolean isContainment() {
		return containmentProperty.isContainment();
	}

	@Override
	public Object getValue(Object dataObject) 
	{	
		if(type.getDataObject() != dataObject)
			throw new RuntimeException("Incorrect data object passed to getValue");
		
		return type.getCollectionElement(this);
	}

	@Override
	public void setValue(Object dataObject, Object propertyValue) 
	{	
		type.setCollectionElement(this, propertyValue);
	}	

	@Override
	public void setCustomAdd(Object dataObject, CollectionEvent event) 
	{
		throw new UnsupportedOperationException("custom add method currently not supported for a Dynamic property");
	}		

	@Override
	public void setCustomRemove(Object dataObject, CollectionEvent event) 
	{
		throw new UnsupportedOperationException("custom remove method currently not supported for a Dynamic property");
	}	

	@Override
	public boolean needsInitialization() {
		return containmentProperty.needsInitialization();
	}

	@Override
	public boolean isComputed() {
		return containmentProperty.isComputed();
	}

	@Override
	public boolean isTransient() {	
		return containmentProperty.isTransient();

	}	

	@Override
	public boolean isBiDirectional() {
		return getOpposite() != null;
	}

	@Override
	public boolean isSymmetricalBiDirectionalType() {
		if(!isBiDirectional())
			return false;

		ExtendedProperty opposite = (ExtendedProperty) getOpposite();

		if(opposite.isMany())
			return this.containmentProperty.getElementType() == opposite.getElementType();
		else
			return this.containmentProperty.getElementType() == opposite.getType();	
	}

	@Override
	public String getMapPath() {
		return containmentProperty.getMapPath();
	}

	@Override
	public boolean isDataType() {
		return containmentProperty.getElementType().isDataType();
	}

	@Override
	public void linkBackPointer(ExtendedDataObject dataObject) {
		ExtendedProperty backPointer = (ExtendedProperty) getOpposite();
		if(backPointer == null)
			return;

		ExtendedDataObject otherSide = (ExtendedDataObject) dataObject.getDataObject(this);
		if(otherSide == null)
			return;

		if(backPointer.isMany()) {
			if(backPointer.isSet()) // Only Set is supported
				((Set) backPointer.getValue(otherSide)).add(dataObject.getInstance());
		} else
			backPointer.setValue(otherSide, dataObject.getInstance());
	}

	@Override
	public void unlinkBackPointer(ExtendedDataObject dataObject) {
		ExtendedProperty backPointer = (ExtendedProperty) getOpposite();
		if(backPointer == null)
			return;

		ExtendedDataObject otherSide = (ExtendedDataObject) dataObject.getDataObject(this);
		if(otherSide == null)
			return;

		if(backPointer.isMany()) {
			if(backPointer.isSet()) // Only Set is supported
				((Set) backPointer.getValue(otherSide)).remove(dataObject.getInstance());
		} else
			backPointer.setValue(otherSide, null);
	}

	@Override
	public String getName() {
		return this.name;
	}

	@Override
	public boolean isMany() {
		return false;
	}

	@Override
	public Object getDefault() {
		return null;
	}

	@Override
	public boolean isReadOnly() {
		return containmentProperty.isReadOnly();
	}

	@Override
	public boolean isNullable() {
		return containmentProperty.isNullable();
	}

	@Override
	public boolean isOpenContent() {
		return true;
	}

	@Override
	public List<?> getInstanceProperties() {
		return type.getProperties();
	}

	@Override
	public Object get(Property property) {
		return null;
	}

	@Override
	public PersistentAttributeType getAssociationType() {
		if(containmentProperty.getAssociationType() == PersistentAttributeType.MANY_TO_MANY)
			return PersistentAttributeType.ONE_TO_MANY;
		else
			return PersistentAttributeType.ONE_TO_ONE;
	}

	@Override
	public boolean isMap() {
		return false;
	}

	@Override
	public boolean isList() {
		return false;
	}

	@Override
	public boolean isSet() {
		return false;
	}

	@Override
	public Object shallowCopy(Object dataObject, ObjectCreator creator) {
		throw new UnsupportedOperationException("Shallow copy not supported for a dynamic property");
	}

	@Override
	public void setMapPath(String mapPath) {
		throw new UnsupportedOperationException("Setting map path is not supported for a dynamic property");
	}
}
