/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

/**
 * SimpleType have no properties
 * 
 * @author Dilip Dalton
 * 
 */
public class DynamicType implements Type {

	private ExtendedDataObject dataObject;
	private Map<String, Property> properties;
	private Map<Property, Object> propertyValues = new HashMap<Property, Object>();

	public DynamicType(ExtendedDataObject dataObject) {

		if(!dataObject.getContainmentProperty().isMany())
			throw new RuntimeException("A DynamicType can only represent a Collection or a Map object");

		this.dataObject = dataObject;
	}

	public ExtendedDataObject getDataObject() {
		return dataObject;
	}

	public Object getCollectionElement(Property property) {
		return propertyValues.get(property);
	}

	public void setCollectionElement(Property property, Object propertyValue) {
		Object instance = dataObject.getInstance();
		if(instance != null) {
			propertyValues.put(property, propertyValue);
			if(Map.class.isAssignableFrom(instance.getClass()))
				((Map)instance).put(property.getName(), propertyValue);
			else if(List.class.isAssignableFrom(instance.getClass()))
				((List)instance).set(Integer.parseInt(property.getName()), propertyValue);
			else if(Set.class.isAssignableFrom(instance.getClass()))
				((Set)instance).add(propertyValue);
		} else 
			throw new RuntimeException("Trying to set/add element on a null collection");
	}	

	@Override
	public String getName() {
		// The name is unique in the namespace because if is qualified by the
		// package name
		return dataObject.getType().getName();
	}

	@Override
	public String getURI() {
		return null;
	}

	@Override
	public Class<?> getInstanceClass() {
		return dataObject.getType().getInstanceClass();
	}

	@Override
	public boolean isInstance(Object object) {
		return dataObject.getType().isInstance(object);
	}

	/**
	 * The property names within the type are unique
	 */
	@Override
	public List<Property> getProperties() {
		if(properties == null) {
			properties = new HashMap<String, Property>();

			Object instance = dataObject.getInstance();
			if(instance != null) {

				// Map
				if(Map.class.isAssignableFrom(instance.getClass())) {
					Map map = (Map) instance;
					for(Object object: map.entrySet()) {
						Map.Entry entry = (Map.Entry) object;
						DynamicProperty property = new DynamicProperty(entry.getKey().toString(), this, (ExtendedProperty) dataObject.getContainmentProperty());
						properties.put(entry.getKey().toString(), property);
						propertyValues.put(property, entry.getValue());
					}
				} else if(List.class.isAssignableFrom(instance.getClass())) {
					List list = (List) instance;
					for(int i = 0; i < list.size(); i++) {
						DynamicProperty property = new DynamicProperty(Integer.toString(i), this, (ExtendedProperty) dataObject.getContainmentProperty());
						properties.put(Integer.toString(i), property);
						propertyValues.put(property, list.get(i));
					}					
				} else if(Set.class.isAssignableFrom(instance.getClass())) {
					Set set = (Set) instance;
					for(Object element: set) {
						String key = dataObject.getCollectionElementId(element);
						if(key == null)
							key = Integer.toString(System.identityHashCode(element));
						DynamicProperty property = new DynamicProperty(key, this, (ExtendedProperty) dataObject.getContainmentProperty());
						properties.put(key, property);
						propertyValues.put(property, element);
					}
				}
			}
		}

		return new ArrayList<Property>(properties.values());
	}

	@Override
	public Property getProperty(String propertyName) {
		return properties.get(propertyName);
	}

	@Override
	public boolean isDataType() {
		return false;
	}

	@Override
	public boolean isOpen() {
		return true;
	}

	@Override
	public boolean isSequenced() {
		if (((ExtendedProperty)dataObject.getContainmentProperty()).isList())
			return true;
		else
			return false;
	}

	@Override
	public boolean isAbstract() {
		return false;
	}

	@Override
	public List<Type> getBaseTypes() {
		return new ArrayList<Type>();
	}

	@Override
	public List<Property> getDeclaredProperties() {
		return new ArrayList<Property>();
	}

	@Override
	public List<?> getAliasNames() {
		return new ArrayList<String>();
	}

	@Override
	public List<?> getInstanceProperties() {
		return getProperties();
	}

	@Override
	public Object get(Property property) {
		return null;
	}
}
