/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.core.EmptyInterceptor;
import ws.ladder.core.Interceptor;
import ws.ladder.model.custom.AssociationStrategy;
import ws.ladder.model.view.ContentView;
import ws.ladder.model.view.Filter;


public class Settings {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	public static final String PATH_DELIMITER = ".";
	public static final String PATH_DELIMITER_REGEX = "\\" + PATH_DELIMITER; // The '.' character is a special character in regular expressions, so it has to be escaped
	public static final String URI_PATH_DELIMITER = "_";	
	public static final String ACTION_OVERRIDE = URI_PATH_DELIMITER + "ACTION" + URI_PATH_DELIMITER;
	public static final String USERKEY_OVERRIDE = URI_PATH_DELIMITER + "USERKEY" + URI_PATH_DELIMITER;	

	// The scope or graph depth of the data that is initialized/retrieved by the traversal 
	protected ContentScope scope = ContentScope.BASE;

	// The amount of the data that is initialized/retrieved by the traversal 
	protected ContentDetail detail = ContentDetail.FULL;	

	// If the scope is ContentScope.VIEW then the actual view is referenced in this property
	protected ContentView view;

	private AggregateAction action = AggregateAction.UPDATE; // specifies the type of action being performed that involves data change in the database

	private Map<String, Object> filters = new HashMap<String, Object>(); // Used for filtering the result

	private Map<String, AggregateAction> actionOverrides; // A collection can have a specific action override set

	private Map<String, String> userkeyOverrides; // An association can be resolved using a userkey	

	private Class<?> desiredSubclass;

	private List<AssociationSetting> associationSettings;

	private Interceptor interceptor = EmptyInterceptor.INSTANCE; // Allow user code to inspect or tweak the processing

	private boolean preFlush;      // Setting this flag to true, flushes the session before it executes the operation
	private boolean postFlush;     // Setting this flag to true, flushes the session after it executes the operation	
	private boolean preClear; // Setting this flag to true clears the session cache before it executes the operation
	private boolean preRefresh;   // Setting this flag to true refreshes the object before it executes the operation	

	private boolean narrow; // Request the query operation to narrow the object to the appropriate type. This can have a performance impact

	private boolean crossAggregate; // Flag to allow an operation to cross aggregate boundaries. Relevant when a view is used.

	private AssociationStrategy associationStrategy;
	
	private List<Filter> additionalFilters = new ArrayList<Filter>();

	private Class<?> narrowedClass;
	
	private List<String> tags = new ArrayList<String>();
	
	private Integer limit;
	private Integer offset;

	public Settings() {
		this.scope = ContentScope.BASE;
		this.detail = ContentDetail.FULL;
		this.action = AggregateAction.UPDATE;
		this.filters = new HashMap<String, Object>();
		this.actionOverrides = new HashMap<String, AggregateAction>();
		this.userkeyOverrides = new HashMap<String, String>();	
		this.associationSettings = new ArrayList<AssociationSetting>();
	}

	public void addTag(String value) {
		tags.add(value);
	}

	public List<String> getTags() {
		return this.tags;
	}

	public Class<?> getNarrowedClass() {
		return narrowedClass;
	}

	public void setNarrowedClass(Class<?> narrowedClass) {
		this.narrowedClass = narrowedClass;
	}

	public AssociationStrategy getAssociationStrategy() {
		return associationStrategy;
	}

	public void setAssociationStrategy(
			AssociationStrategy associationStrategy) {
		this.associationStrategy = associationStrategy;
	}	

	public void addAssociation(AssociationSetting associationSetting) {
		this.associationSettings.add(associationSetting);
	}

	public List<AssociationSetting> getAssociationSettings() {
		return this.associationSettings;
	}

	public void init(ContentScope scope, ContentDetail detail, Map<String, String> queryParams) {
		init(scope, detail, null, queryParams);
	}

	public void init(ContentView view, Map<String, String> queryParams) {
		init(ContentScope.VIEW, null, view, queryParams);
	}	

	public void init(ContentScope scope, ContentDetail detail, ContentView view, Map<String, String> queryParams) {

		this.scope = scope;
		this.detail = detail;
		this.view = view;
		this.filters = populateFilters(queryParams);
		this.actionOverrides = getActionOverrides(queryParams);
		this.userkeyOverrides = getUserkeyOverrides(queryParams);

		initMutableAction(queryParams);
	}

	public Map<String, Object> populateFilters(Map<String, String> queryParams) {
		if(view == null || queryParams == null)
			return filters;

		for(String key: view.getAttributeList()) {
			String encodedKey = encodeParam(key);
			if(queryParams.containsKey(encodedKey))
				filters.put(key, queryParams.get(encodedKey));					
		}

		return filters;
	}

	protected void initMutableAction(Map<String, String> queryParams) {
		if(queryParams == null)
			return;

		for(Entry<String, String> entry: queryParams.entrySet()) {
			String key = entry.getKey();
			String value = entry.getValue();
			if(key.toUpperCase().equals(ACTION_OVERRIDE.toUpperCase())) {
				action = AggregateAction.valueOf(value.toUpperCase());
				break;
			}
		}

		if(action == null)
			throw new IllegalArgumentException("Action has an invalid value should be one of the following values - CREATE, UPDATE, MERGE.");
	}

	// Currently supports only VIEW scope. Can probably expand to other scopes.
	public Map<String, AggregateAction> getActionOverrides(Map<String, String> queryParams) {

		if(queryParams == null)
			return new HashMap<String, AggregateAction>();

		Map<String, AggregateAction> result = new HashMap<String, AggregateAction>();
		for(Entry<String, String> entry: queryParams.entrySet()) {
			String key = entry.getKey();
			String mutableAction = entry.getValue();
			String toUpperKey = key.toUpperCase();
			if(!toUpperKey.endsWith(ACTION_OVERRIDE))
				continue;

			int endIndex = toUpperKey.indexOf(ACTION_OVERRIDE);
			String propertyPathSuffix = toUpperKey.substring(0, endIndex);
			AggregateAction actionOverride = AggregateAction.valueOf(mutableAction.toUpperCase());			
			result.put(decodeParam(propertyPathSuffix), actionOverride);
		}		


		return result;
	}

	public boolean doMerge(CallInfo ci) {
		return doAction(ci, AggregateAction.MERGE);
	}	

	public boolean doUpdate(CallInfo ci) {
		return doAction(ci, AggregateAction.UPDATE);
	}	

	public boolean doAction(CallInfo ci, AggregateAction aggregateAction) {
		if(!(actionOverrides == null || actionOverrides.isEmpty() || ci.getInputPropertyPath() == null) ){
			if(actionOverrides.get(ci.getInputPropertyPath().toUpperCase()) != null){
				if (actionOverrides.get(ci.getInputPropertyPath().toUpperCase()) == aggregateAction){
					return true;
				}else{
					return false;
				}				
			}
		}
		if(action == aggregateAction)
			return true;

		return false;
	}	

	/**
	 * The key name does not have to refer to the full property path. 
	 * For e.g., resourceBinding.resourceProvider_userkey_=name
	 * Since this property can be referenced from ServiceComponent at different levels
	 * 
	 * @param propertyPath
	 * @return
	 */
	public String getUserKeyOverride(String propertyPath) {
		// Get the first match
		for(String key: userkeyOverrides.keySet()) {
			if(propertyPath.toUpperCase().endsWith(key)) {
				logger.info("Got key override, property path: " + propertyPath + ", matched userKey: " + key + ", userkey value: " + userkeyOverrides.get(key));
				return userkeyOverrides.get(key);
			}
		}

		return null;
	}

	public Map<String, String> getUserkeyOverrides(Map<String, String> queryParams) {
		if(queryParams == null)
			return new HashMap<String, String>();

		Map<String, String> toUpperParams = new HashMap<String, String>();
		for(String key: queryParams.keySet())
			toUpperParams.put(key.toUpperCase(), queryParams.get(key));

		Map<String, String> result = new HashMap<String, String>();
		for(String key: toUpperParams.keySet()) {
			if(!key.endsWith(USERKEY_OVERRIDE))
				continue;

			int endIndex = key.indexOf(USERKEY_OVERRIDE);
			String propertyPathSuffix = key.substring(0, endIndex);
			result.put(decodeParam(propertyPathSuffix), toUpperParams.get(key));
		}

		return result;
	}		

	public static String decodeParam(String param) {
		return param.replace(URI_PATH_DELIMITER, PATH_DELIMITER);
	}

	public static String encodeParam(String param) {
		return param.replace(PATH_DELIMITER, URI_PATH_DELIMITER);
	}			

	public ContentScope getScope() {
		return scope;
	}

	public void setScope(ContentScope scope) {
		this.scope = scope;
	}

	public ContentDetail getDetail() {
		return detail;
	}

	public void setDetail(ContentDetail detail) {
		this.detail = detail;
	}

	public ContentView getView() {
		return view;
	}

	public void setView(ContentView view) {
		this.view = view;
	}

	public AggregateAction getAction() {
		return action;
	}

	public void setAction(AggregateAction mutableAction) {
		// this.mutableAction can never by null
		if(mutableAction != null)
			this.action = mutableAction;
	}

	public Map<String, Object> getFilters() {
		return filters;
	}

	public void setFilters(Map<String, Object> filters) {
		this.filters = filters;
	}

	public Map<String, AggregateAction> getActionOverrides() {
		return actionOverrides;
	}

	public void setActionOverrides(Map<String, AggregateAction> overrides) {
		this.actionOverrides = overrides;
	}

	public Map<String, String> getUserkeyOverrides() {
		return userkeyOverrides;
	}

	public Class<?> getDesiredSubclass() {
		return desiredSubclass;
	}

	public void setDesiredSubclass(Class<?> desiredSubclass) {
		this.desiredSubclass = desiredSubclass;
	}

	public Interceptor getInterceptor() {
		return this.interceptor;
	}

	public void setInterceptor(Interceptor interceptor) {
		this.interceptor = interceptor;
	}

	public boolean doPreFlush() {
		return preFlush;
	}

	public void setPreFlush(boolean flush) {
		this.preFlush = flush;
	}

	public boolean doPostFlush() {
		return postFlush;
	}	

	public void setPostFlush(boolean flush) {
		this.postFlush = flush;
	}	
	
	public void setPreClear(boolean value) {
		this.preClear = value;
	}
	
	public boolean doPreClear() {
		return this.preClear;
	}
	
	public void setPreRefresh(boolean value) {
		this.preRefresh = value;
	}
	
	public boolean doPreRefresh() {
		return this.preRefresh;
	}	

	public boolean doNarrow() {
		return narrow;
	}

	public void setNarrow(boolean narrow) {
		this.narrow = narrow;
	}

	public boolean permitCrossAggregate() {
		return crossAggregate;
	}

	public void setCrossAggregate(boolean crossAggregate) {
		this.crossAggregate = crossAggregate;
	}

	public void addFilter(String name, Object value) {
		filters.put(name, value);
	}

	public void initNarrowClass(TypeNarrower typeNarrower, Object entity, TypeMapper typeMapper)
	{
		if(getView() == null)
			return;
		
		narrowedClass = typeNarrower.narrow(entity, getView().getName());
		narrowedClass = typeMapper.toReference(narrowedClass);
	}

	public List<Filter> getAdditionalFilters() {
		return this.additionalFilters;
	}
	
	public void addFilter(String filterExpression) {
		Filter newFilter = new Filter();
		newFilter.setExpression(filterExpression);
		
		this.additionalFilters.add(newFilter);
	}

	public Integer getOffset() {
		return offset;
	}

	public void setOffset(Integer offset) {
		this.offset = offset;
	}

	public Integer getLimit() {
		return limit;
	}

	public void setLimit(Integer limit) {
		this.limit = limit;
	}
}
