/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.action;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;


public abstract class CollectionUpdateAction implements Executable {
	protected Set<AddElementAction> addByOppositeActions = new HashSet<AddElementAction>();
	protected Set<RemoveElementAction> removeByOppositeActions = new HashSet<RemoveElementAction>();	

	public abstract void addAction(Executable action);

	public abstract void linkElement(CallInfo callInfo, String dynamicPropertyName, boolean isNew) throws Exception;

	public ExtendedDataObject getTriggerObject() {
		throw new UnsupportedOperationException("It is not meaningful to call this method on a CollectionUpdateAction since a collection can be updated from multiple sources and does not represent a single execution");
	}

	public void addTriggeredByOppositeAction(ElementAction action) {
		if(AddElementAction.class.isAssignableFrom(action.getClass()))
			addByOppositeActions.add((AddElementAction) action);
		else if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
			removeByOppositeActions.add((RemoveElementAction) action);
	}

	@Override
	public Executable getTriggeringAction() {
		return null;
	}

	public void linkElements(CallInfo callInfo) throws Exception {
		ExtendedDataObject input  = ((ExtendedDataObject)callInfo.getInput());
		ExtendedDataObject output = ((ExtendedDataObject)callInfo.getOutput());

		Map outputMap = new HashMap<Object, ExtendedDataObject>();
		for(ExtendedDataObject element: output.getList())
			if(element.getCollectionElementId() != null)
				outputMap.put(element.getCollectionElementId(), element);		

		Set<String> inputIds = new HashSet<String>();
		if(callInfo.getInput() != null) {
			List<ExtendedDataObject> inputList = input.getList();
			for(int i = 0; i < inputList.size(); i++) {
				ExtendedDataObject element = inputList.get(i);
				String key = element.getCollectionElementId();
				if(key != null)
					inputIds.add(key);
			}

			Map<Object, Set<String>> elementKeys = getElementKeysMap(input);
			for(ExtendedDataObject sourceElement: inputList) {
				Set<String> keys = elementKeys.get(sourceElement.getInstance());			
				String id = sourceElement.getCollectionElementId();

				boolean isNew = id == null || !outputMap.containsKey(id);

				if(keys != null)
					for(String key: keys)
						processLink(key, sourceElement, callInfo, isNew);
				else
					processLink(null, sourceElement, callInfo, isNew);					
			}
		}
	}
	
	private void processLink(String key, ExtendedDataObject sourceElement, CallInfo callInfo, boolean isNew) throws Exception {
		CallInfo next = new CallInfo(sourceElement, null, callInfo, null);
		next.setOutput(callInfo.getOutputObjectCreator().createTarget(next, null, null));
		linkElement(next, key, isNew);		
	}

	public Map<Object, Set<String>> getElementKeysMap(ExtendedDataObject input) {
		Collection collection = (Collection) input.getInstance();

		Map<Object, Set<String>> result = new HashMap<Object, Set<String>>();
		for(Object object: collection)
			result.put(object, null);

		return result;
	}

	public void unlinkElements(CallInfo callInfo) {

		/* algorithm
		 * 1. All the objects in the targetCollection should have ids since these are entities
		 * 2. We get a list of all ids from the sourceCollection
		 * 3. All ids that are not present in the sourceCollection are obsolete and needs to be removed
		 * 
		 * For a Map, we only honor entity in the value and not in the key. So we just need to check the values collection
		 * of the source and target maps.
		 */		
		ExtendedDataObject input  = ((ExtendedDataObject)callInfo.getInput());
		ExtendedDataObject output = ((ExtendedDataObject)callInfo.getOutput());

		Map targetKeys = new HashMap<Object, Object>();
		for(ExtendedDataObject element: output.getList())
			targetKeys.put(element.getCollectionElementId(), element);		

		Map sourceKeys = new HashMap<Object, Object>();		
		if(callInfo.getInput() != null) 
			for(ExtendedDataObject element: input.getList()) {
				Object idValue = element.getCollectionElementId();
				if(idValue != null)
					sourceKeys.put(idValue, element);
			}			

		for(Object entry : targetKeys.entrySet()) {
			Object targetKey = ((Map.Entry) entry).getKey();
			Object targetElement = ((Map.Entry) entry).getValue();

			if(!sourceKeys.containsKey(targetKey)) { // obsolete element being marked for removal
				CallInfo next = new CallInfo(null, targetElement, callInfo, null);
				next.setOutput(callInfo.getOutputObjectCreator().createTarget(next, targetElement, null));
				unlinkElement(next, (ExtendedDataObject) next.getOutput(), targetKey);
			}
		}		

	}

	protected void unlinkElement(CallInfo callInfo, ExtendedDataObject collectionElement, Object position) {
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getOutput();

		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getOutput(), callInfo.getParent().getOutputProperty());
		Executable originalAction = new RemoveElementAction(propertyKey, collectionElement, collectionOwner, null, position, callInfo);
		this.addAction(originalAction);		

		// Set the backRef value to null. This is an asynchronous activity
		PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getOutputProperty().getOpposite());			
		callInfo.getOutputRoot().getDependencyGraph().addAction(new SetterAction(null, newOppositeKey, originalAction)); 	

	}		
}
