/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.action;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;

import ws.ladder.model.AggregateAction;
import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.event.CollectionElement;
import ws.ladder.model.event.CollectionEvent;
import ws.ladder.model.operation.AbstractOperation;
import ws.ladder.util.ClassUtil;


public class ListUpdateAction extends CollectionUpdateAction {

	private PropertyKey key;
	private List<AddElementAction>        addActions = new ArrayList<AddElementAction>();
	private List<RemoveElementAction>     removeActions = new ArrayList<RemoveElementAction>();

	public ListUpdateAction(PropertyKey key) {
		this.key = key;
	}

	@Override
	public void execute() {
		boolean collectionModified = false;

		// TODO: remove elements

		// Sort by position
		Map<Integer, AddElementAction> listOrderedActions = new TreeMap<Integer, AddElementAction>();

		for(AddElementAction action: addActions) 
			listOrderedActions.put(Integer.parseInt(action.getPosition().toString()), action);
		// TODO: Handler conflicts
		for(AddElementAction action: addByOppositeActions) 
			listOrderedActions.put(Integer.parseInt(action.getPosition().toString()), action);		

		for(AddElementAction action: listOrderedActions.values()) {
			// Add the element to the List
			customAddToList(action);		
			collectionModified = true;			
		}

		// checkOutOfSync();

		if(collectionModified)
			AbstractOperation.setModified(key.getDataObject());		
	}		

	@Override
	public Map<Object, Set<String>> getElementKeysMap(ExtendedDataObject input) {
		List list = (List) input.getInstance();

		Map<Object, Set<String>> result = new HashMap<Object, Set<String>>();
		for(int i = 0; i < list.size(); i++) {
			Object object = list.get(i);
			Set<String> keys = result.get(object);
			if(keys == null) {
				keys = new HashSet<String>();
				result.put(object, keys);
			}
			keys.add(Integer.toString(i));
		}

		return result;
	}	

	@Override
	public void addAction(Executable action) {
		if(AddElementAction.class.isAssignableFrom(action.getClass()))
			addActions.add((AddElementAction) action);
		else if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
			removeActions.add((RemoveElementAction) action);	
		else
			throw new IllegalArgumentException("A list can have only add, remove actions");
	}

	@Override
	public PropertyKey getKey() {
		return key;
	}

	public void unlinkElements(CallInfo callInfo) {
		super.unlinkElements(callInfo);

		ExtendedDataObject input  = ((ExtendedDataObject)callInfo.getInput());
		ExtendedDataObject output = ((ExtendedDataObject)callInfo.getOutput());

		if(output == null)
			return;

		List inputList = (List) input.getInstance();
		List outputList = (List) output.getInstance();

		if(outputList.size() > inputList.size()) // truncate the remaining portion
			truncateToSize(outputList, inputList.size());
	}

	private void truncateToSize(List outputList, int toSize) {
		assert(outputList.size() >= toSize);

		for(int i = outputList.size(); i > toSize; i--) 
			outputList.remove(i-1);
	}

	public void linkElement(CallInfo next, String dynamicProperty, boolean isNew) throws Exception {
		AbstractOperation oper = (AbstractOperation) next.getOperation();
		ExtendedDataObject targetElement = (ExtendedDataObject) next.getOutput();
		if(next.getSettings().getAction() != AggregateAction.CREATE && targetElement != null) // updating an existing object
			next.setOutput(targetElement);
		else
			next.setOutput(oper.createTarget(next, null));

		oper.processValue(next);

		// AddElementAction
		int i = Integer.valueOf(dynamicProperty);
		addListElement(next, (ExtendedDataObject) next.getOutput(), i);

		// Add edge in dependency graph
		AbstractOperation.addEdge(next);	
	}

	protected void addListElement(CallInfo callInfo, ExtendedDataObject collectionElement, Object position) {
		Object collection = callInfo.getParent().getOutput();
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getOutput();

		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) collection, callInfo.getParent().getOutputProperty());
		Executable originalAction = new AddElementAction(propertyKey, collectionElement, collectionOwner, null, position, callInfo);
		this.addAction(originalAction);		

		// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
		// TODO: handle MANY_TO_MANY backRef
		if(callInfo.getParent().getOutputProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getOutputProperty().getOpposite());			
			callInfo.getOutputRoot().getDependencyGraph().addAction(new SetterAction(collectionOwner, newOppositeKey, originalAction));
		}
	}

	protected void customAddToList(AddElementAction action) {
		CallInfo callInfo = action.getCallInfo();

		ExtendedDataObject collectionOwner  = action.getCollectionOwner();
		Object collection                   = ((ExtendedProperty) action.getKey().getProperty()).getValue(collectionOwner);
		Object collectionElement            = action.getCollectionElement();
		ExtendedProperty collectionProperty = (ExtendedProperty) action.getKey().getProperty();

		// It is possible for the collection object to be null, if this is due to the MANY_TO_ONE end only being set
		try {
			if(collection == null) {
				ExtendedDataObject collectionDataObject = (ExtendedDataObject) collectionOwner.createDataObject((Object)null, collectionProperty.getType());
				collection = collectionDataObject.getInstance();
				((ExtendedProperty) action.getKey().getProperty()).setValue(collectionOwner, collection);
			} 
		} catch(Exception e) {
			throw ClassUtil.wrapRun(e);
		}


		// check if there is a custom collection add
		Method customCollectionAdd = ClassUtil.getCustomAdd(collectionProperty.getElementType().getInstanceClass(), collectionProperty.getName());
		if(customCollectionAdd != null) {
			CollectionEvent event = new CollectionElement(callInfo, collection, collectionElement);
			collectionProperty.setCustomAdd(collection, event);
		} else {
			if(((java.util.Collection<Object>)collection).size() == Integer.parseInt(action.getPosition().toString()) )
				((java.util.Collection<Object>)collection).add(ClassUtil.getInstance(collectionElement));
			else
				throw new IllegalArgumentException("Trying to add a list element at the wrong position: " + action.getPosition() + ", list size: " + 
						((java.util.Collection<Object>)collection).size());
		}
	}		
}
