/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.action;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ws.ladder.model.AggregateAction;
import ws.ladder.model.CallInfo;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.event.CollectionElement;
import ws.ladder.model.event.CollectionEvent;
import ws.ladder.model.operation.AbstractOperation;
import ws.ladder.util.ClassUtil;


public class MapUpdateAction extends CollectionUpdateAction {

	private PropertyKey key;
	private List<AddElementAction>        addActions = new ArrayList<AddElementAction>();
	private List<RemoveElementAction>     removeActions = new ArrayList<RemoveElementAction>();

	public MapUpdateAction(PropertyKey key) {
		this.key = key;
	}

	@Override
	public void execute() {


		// TODO: Invoke RemoveElementAction, AddElementAction and RepositionElementAction
		// Any incomplete actions need to be resolved or an error raised
		// An action is incomplete 
		// 1. if the position is null or
		// 2. position is not null (duplicates will not be created) and the collection is a list or a map
		// 3. setModified flag on map data object

		for(AddElementAction action: addActions) {
			// Add the element to the Map
			customAddToMap(action);		
		}

	}

	@Override
	public void addAction(Executable action) {
		if(AddElementAction.class.isAssignableFrom(action.getClass()))
			addActions.add((AddElementAction) action);
		else if(RemoveElementAction.class.isAssignableFrom(action.getClass()))
			removeActions.add((RemoveElementAction) action);	
		else
			throw new IllegalArgumentException("A Map can have only add, remove actions");
	}
	
	@Override
	public Map<Object, Set<String>> getElementKeysMap(ExtendedDataObject input) {
		Map map = (Map) input.getInstance();

		Map<Object, Set<String>> result = new HashMap<Object, Set<String>>();
		for(Object item: map.entrySet()) {
			Map.Entry entry = (Map.Entry) item;
			Set<String> keys = result.get(entry.getValue());
			if(keys == null) {
				keys = new HashSet<String>();
				result.put(entry.getValue(), keys);
			}
			keys.add(entry.getKey().toString());
		}

		return result;
	}	

	@Override
	public PropertyKey getKey() {
		return key;
	}

	public void linkElement(CallInfo next, String dynamicProperty, boolean isNew) throws Exception {
		AbstractOperation oper = (AbstractOperation) next.getOperation();
		ExtendedDataObject targetElement = (ExtendedDataObject) next.getOutput();
		if(next.getSettings().getAction() != AggregateAction.CREATE && targetElement != null)// updating an existing object
			next.setOutput(targetElement);
		else
			next.setOutput(oper.createTarget(next, null));
		oper.processValue(next);

		// AddElementAction
		addMapElement(next, (ExtendedDataObject) next.getOutput(), dynamicProperty);			

		// Add edge in dependency graph
		AbstractOperation.addEdge(next);	
	}	

	protected void addMapElement(CallInfo callInfo, ExtendedDataObject collectionElement, String position) {
		ExtendedDataObject collectionOwner = (ExtendedDataObject) callInfo.getParent().getParent().getOutput();

		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getOutput(), callInfo.getParent().getOutputProperty());
		AddElementAction originalAction = new AddElementAction(propertyKey, collectionElement, collectionOwner, null, position, callInfo);
		this.addAction(originalAction);

		// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
		// TODO: handle MANY_TO_MANY backRef			
		if(originalAction != null && callInfo.getParent().getOutputProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey(collectionElement, callInfo.getParent().getOutputProperty().getOpposite());			
			callInfo.getOutputRoot().getDependencyGraph().addAction(new SetterAction(collectionElement, newOppositeKey, originalAction));
		}
	}	
	
	protected void customAddToMap(AddElementAction action) {
		CallInfo callInfo = action.getCallInfo();

		ExtendedDataObject collectionOwner  = action.getCollectionOwner();
		Object collection                   = ((ExtendedProperty) action.getKey().getProperty()).getValue(collectionOwner);
		Object collectionElement            = action.getCollectionElement();
		ExtendedProperty collectionProperty = (ExtendedProperty) action.getKey().getProperty();

		// It is possible for the collection object to be null, if this is due to the MANY_TO_ONE end only being set
		try {
			if(collection == null) {
				ExtendedDataObject collectionDataObject = (ExtendedDataObject) collectionOwner.createDataObject((Object)null, collectionProperty.getType());
				collection = collectionDataObject.getInstance();
				((ExtendedProperty) action.getKey().getProperty()).setValue(collectionOwner, collection);
			} 
		} catch(Exception e) {
			throw ClassUtil.wrapRun(e);
		}

		// check if there is a custom collection add
		Method customCollectionAdd = ClassUtil.getCustomAdd(collectionProperty.getElementType().getInstanceClass(), collectionProperty.getName());
		if(customCollectionAdd != null) {
			CollectionEvent event = new CollectionElement(callInfo, collection, collectionElement);
			collectionProperty.setCustomAdd(collection, event);
		} else {
			((java.util.Map)collection).put(action.getPosition().toString(), ClassUtil.getInstance(collectionElement));
		}
	}			
}
