/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.model.operation;

import javax.persistence.metamodel.Attribute.PersistentAttributeType;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import ws.ladder.exception.PropertyMappingException;
import ws.ladder.model.CallInfo;
import ws.ladder.model.DynamicProperty;
import ws.ladder.model.ExtendedDataObject;
import ws.ladder.model.ExtendedProperty;
import ws.ladder.model.ProcessingStage;
import ws.ladder.model.action.AddElementAction;
import ws.ladder.model.action.CollectionUpdateAction;
import ws.ladder.model.action.Executable;
import ws.ladder.model.action.PropertyKey;
import ws.ladder.model.action.RemoveElementAction;
import ws.ladder.model.action.SetterAction;


public class ModifyOperation extends AbstractOperation {
	private static final Logger logger = LogManager.getLogger(new Exception().getStackTrace()[0].getClassName());

	private Object result;

	public void removeObsoleteEntities(CallInfo callInfo, CollectionUpdateAction migratorAction) {

		// If we are merging then obsolete elements are not removed 		
		if(callInfo.getSettings().doMerge(callInfo))
			return;		

		migratorAction.unlinkElements(callInfo);
	}	

	@Override
	public void processCollection(CallInfo callInfo) throws Exception {

		if(!callInfo.isCascadable()) // Hibernate will not update this collection
			return;

		if(callInfo.getStage() != ProcessingStage.UPDATE)
			return;		

		updateToMany(callInfo, null);
	}		

	protected void updateToMany(CallInfo callInfo, Executable triggeringAction) throws Exception {
		// Update the collection/map. This includes adding/removing/repositioning elements
		PropertyKey propertyKey = new PropertyKey((ExtendedDataObject) callInfo.getParent().getOutput(), callInfo.getOutputProperty());
		CollectionUpdateAction migratorAction = callInfo.getOutputRoot().getDependencyGraph().getOrCreateMigratorAction(propertyKey);

		removeObsoleteEntities(callInfo, migratorAction);
		migratorAction.linkElements(callInfo);
		//migratorAction.addOrReposition(callInfo);
	}

	@Override
	protected boolean shouldUnlink(CallInfo ci) {

		if(ci.getInput() == null || ci.getInputProperty() == null)
			return true;

		return ci.getInputProperty().isTransient();
	}	

	@Override
	protected void processNullValue(CallInfo ci) throws Exception {

		if(ci.getSettings().doUpdate(ci) && ci.getStage() == ProcessingStage.UPDATE) {

			if(ci.getInputProperty() != null && ci.getInputProperty().isMany()) {
				ci.setOutput(createTarget(ci, ci.getOutputFromParent(), null)); // Get the persistent data object into target
				processCollection(ci);
			} else
				unlinkToOne(ci);
		}
	}	

	protected void unlinkToOne(CallInfo ci) {
		if(ci.getOutputProperty().getAssociationType() != PersistentAttributeType.ONE_TO_ONE &&
				ci.getOutputProperty().getAssociationType() != PersistentAttributeType.MANY_TO_ONE)
			return;
		
		// Unlinking for dynamic properties is handled by remove obsolete method
		if(DynamicProperty.class.isAssignableFrom(ci.getOutputProperty().getClass()))
			return;

		if(ci.getOutputFromParent() == null)
			return;		

		// Set the target to null
		PropertyKey key = new PropertyKey((ExtendedDataObject)ci.getParentOutputEntity(), ci.getOutputProperty());		
		Executable originalAction = new SetterAction(null, key, null);
		ci.getOutputRoot().getDependencyGraph().addAction(originalAction);		

		// bi-directional link
		if(ci.getOutputProperty().isBiDirectional()) {

			ci.setOutput(createTarget(ci, ci.getOutputFromParent(), null)); // Get the persistent data object into target
			PropertyKey oppositeKey = new PropertyKey((ExtendedDataObject)ci.getOutput(), ci.getOutputProperty().getOpposite());

			// set the backRef from the opposite object to null	
			Executable oppositeAction = new SetterAction(null, oppositeKey, originalAction);
			if(ci.getOutputProperty().getAssociationType() == PersistentAttributeType.MANY_TO_ONE) {
				Object position = (oppositeKey.getProperty() == null) ? null : ((ExtendedProperty)oppositeKey.getProperty()).getValue(key.getDataObject());
				oppositeAction = new RemoveElementAction(oppositeKey, key.getDataObject(), (ExtendedDataObject) ci.getOutput(), originalAction, position, ci);				
			}
			ci.getOutputRoot().getDependencyGraph().addAction(oppositeAction);			
		}
	}

	private boolean isSameTarget(CallInfo ci, Object value) {
		return ci.getOutputFromParent() == ((ExtendedDataObject)value).getInstance();
	}

	protected void linkToOne(CallInfo ci, Object value) {
		if(ci.getOutputProperty().getAssociationType() != PersistentAttributeType.ONE_TO_ONE &&
				ci.getOutputProperty().getAssociationType() != PersistentAttributeType.MANY_TO_ONE)
			return;

		// If this is a uni-directional OneToOne, we just have to set the target to value
		PropertyKey key            = new PropertyKey((ExtendedDataObject)ci.getParentOutputEntity(), ci.getOutputProperty());
		Executable originalAction  = new SetterAction(value, key, null);

		// Set the target to value
		if(!isSameTarget(ci, value))		
			ci.getOutputRoot().getDependencyGraph().addAction(originalAction);			

		if(ci.getOutputProperty().isBiDirectional()) {
			PropertyKey newOppositeKey = new PropertyKey((ExtendedDataObject)value, ci.getOutputProperty().getOpposite());	

			// If the target has an existing opposite object, its backRef needs to be set to null 
			if(ci.getOutputFromParent() != null && !isSameTarget(ci, value) && !DynamicProperty.class.isAssignableFrom(ci.getOutputProperty().getClass())) {
				CallInfo newCall = new CallInfo(null, null, ci, ci.getOutputProperty());
				ExtendedDataObject oldObject = (ExtendedDataObject) createTarget(newCall, ci.getOutputFromParent(), null); // Get the persistent data object into target

				// Set the backRef of the opposite object to null
				PropertyKey oldOppositeKey = new PropertyKey(oldObject, ci.getOutputProperty().getOpposite());

				// Set the backRef from the old value to null
				Executable oppositeAction = new SetterAction(null, oldOppositeKey, originalAction);				
				if(ci.getOutputProperty().getAssociationType() == PersistentAttributeType.MANY_TO_ONE) {
					Object position = getPosition((ExtendedProperty) oldOppositeKey.getProperty(), key.getDataObject());
					oppositeAction = new RemoveElementAction(oldOppositeKey, key.getDataObject(), oldObject, originalAction, position, ci);				
				}
				ci.getOutputRoot().getDependencyGraph().addAction(oppositeAction);	
			}			 

			// Set the backRef of value to point to the parent target (or to the ancester in case of embedded property)
			Executable oppositeAction = new SetterAction(ci.getParentOutputEntity(), newOppositeKey, originalAction);			
			if(ci.getOutputProperty().getAssociationType() == PersistentAttributeType.MANY_TO_ONE) {
				Object position = getPosition((ExtendedProperty) newOppositeKey.getProperty(), key.getDataObject());
				oppositeAction = new AddElementAction(newOppositeKey, key.getDataObject(), (ExtendedDataObject) value, originalAction, position, ci);				
			}
			ci.getOutputRoot().getDependencyGraph().addAction(oppositeAction);		
		}
	}
	
	private Object getPosition(ExtendedProperty property, ExtendedDataObject collectionElement) {
		if(property == null)
			return null;
		
		if(!property.isDataType()) {
			if(property.getPositionProperty() != null)
				return ((ExtendedProperty)property.getPositionProperty()).getValue(collectionElement);
		}
		
		return null;
	}

	protected Object setPropertyTarget(CallInfo ci, Object propertyTarget) throws PropertyMappingException {
		if(ci.getStage() != ProcessingStage.UPDATE)
			return propertyTarget;

		// TODO: If the property is a mapKey or an indexKey then the toMany side needs to know about the reposition action 

		if(ci.getOutputProperty().isMany())
			super.setPropertyTarget(ci, propertyTarget);
		else 
			linkToOne(ci, propertyTarget);

		return propertyTarget;
	}

	@Override
	protected void setResult(Object target) {
		this.result = target;
	}	

	@Override
	public Object getResult() {
		return result;
	}	
}
