/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.service;

import java.io.Serializable;
import java.util.List;
import java.util.Map;

import javax.inject.Inject;
import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;

import ws.ladder.model.Type;
import ws.ladder.model.view.JPAQuery;
import ws.ladder.model.view.Query;

public class JPAPersistenceOrchestrator extends AbstractPersistenceOrchestrator {
	@PersistenceContext
	private EntityManager entityManager;
	
    @Inject
    EntityManagerFactory entityManagerFactory;
	
	/**
	 * Save the entity in the persistence store
	 * 
	 * @param entity
	 * @return
	 */
	@Override
	public void saveOrUpdate(Object entity) {
		entityManager.persist(entity);		
	}
	
	@Override 
	public void clear() {
		entityManager.clear();
	}
	
	@Override 
	public void refresh(Object object) {
		entityManager.refresh(object);
	}		

	/**
	 * Delete the entity from the persistence store
	 * 
	 * @param entity
	 */
	@Override
	public void delete(Object entity) {
		entityManager.remove(entity);		
	}

	@Override
	public void flush() {
		entityManager.flush();
	}

	@Override
	public Object findById(Class<?> persistentClass, Object id) {
	   return entityManager.find(persistentClass, (Serializable) id);
	}
	
	@Override
	public Object findByProperty(Type type, Map<String, String> propertyValues) {
		CriteriaBuilder builder = entityManagerFactory.getCriteriaBuilder();
		CriteriaQuery<Object> crit = builder.createQuery();
		
		Class<?> persistentClass = type.getInstanceClass();
		Root from = crit.from(persistentClass);
		CriteriaQuery<Object> select = crit.select(from);

		Predicate[] predicates = new Predicate[propertyValues.size()];
		int index = 0;
		for(String property: propertyValues.keySet())
			predicates[index++] = builder.equal(from.get(property), propertyValues.get(property));

		crit.where(builder.and(predicates));

		TypedQuery<Object> typedQuery = entityManager.createQuery(select);
		List<Object> resultList = typedQuery.getResultList();
		if (resultList != null && !resultList.isEmpty()) {
			return resultList.get(0);
		}
		return null;		
	}	

	@Override
	public QueryCapability getQueryCapability() {
		return new JPAQueryCapability();
	}


	@Override
	public Query getQuery(String queryString, boolean isSQL) {
		
		Query result = null;
		if(isSQL) {
			result = new JPAQuery(entityManager.createNativeQuery(queryString));
		} else
			result = new JPAQuery(entityManager.createQuery(queryString));

		return result;
	}		
}