/**
 * Ladder, empowering Model Driven Architecture in J2EE applications
 *
 * Copyright (c) 2012, Dilip Dalton
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *  http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT 
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *
 * See the License for the specific language governing permissions and limitations 
 * under the License.
 */

package ws.ladder.service;

import java.util.Map;

import ws.ladder.model.CallInfo;
import ws.ladder.model.Type;
import ws.ladder.model.TypeMapper;
import ws.ladder.model.view.Query;

public interface PersistenceOrchestrator {
	
	/**
	 * Find the entity by its primary key
	 * @param id
	 * @return
	 */
	public Object findById(Class<?> persistentClass, Object id);	
	
	/**
	 * Find the entity by one or more property values
	 * @param id
	 * @return
	 */
	public Object findByProperty(Type type, Map<String, String> propertyValues);		

    /**
     * Save the entity in the persistence store
     * @param entity
     * @return
     */
    public void saveOrUpdate(Object entity);
    
    /**
     * Delete the entity from the persistence store
     * @param entity
     */
    public void delete(Object entity);
    
    /**
     * Flush any buffered content to the persistence store
     */
    public void flush();
    
    /**
     * Clears any objects held in the persistence cache
     */
    public void clear();
    
    /**
     * Refreshes the object held in the persistence cache
     */
    public void refresh(Object object);    
    
    /**
     * Indicates if this persistence orchestrator automatically tracks version 
     * If not, the framework takes care of updating the version in order to support optimistic concurrency control
     */
    public boolean supportsVersionTracking();
    
    /**
     * Returns the mechanisms supported for querying
     * @return
     */
    public QueryCapability getQueryCapability();
    
    /**
     * Some ORMs cannot resolve all the issues while persisting an aggregate. 
     * For e.g., cycle dependencies within an aggregate is a problem while persisting it.
     * 
     * This is usually not a problem with NoSQL databases.
     * Setting this flag to false, allows the framework to individually persist
     * objects within the aggregate in based on topological ordering of the required fields.
     */
    public boolean canProcessAggregate();
    
	/**
	 * Check if there is an existing persistent instance with the identifier 
	 * specified in object
	 * 
	 * @param object
	 * @return A persistent object or null 
	 */
	public Object getPersistentObject(CallInfo callInfo, TypeMapper typeMapper);	
	
	/**
	 * Return the query object specific to the persistence mechanism
	 */
	public Query getQuery(String queryString, boolean isSQL);
}